@@ -13,24 +13,27 @@
 ##############################################################################
 """Test execution
 """
+from __future__ import print_function
 
 import subprocess
-
-import cStringIO
 import errno
 import gc
-import Queue
 import re
 import sys
 import threading
 import time
 import traceback
-import unittest
+try:
+    import unittest2 as unittest
+except ImportError:
+    import unittest
 
+from six import StringIO
 from zope.testrunner.find import import_name
 from zope.testrunner.find import name_from_layer, _layer_name_cache
 from zope.testrunner.refcount import TrackRefs
 from zope.testrunner.options import get_options
+import zope.testrunner
 import zope.testrunner.coverage
 import zope.testrunner._doctest
 import zope.testrunner.logsupport
@@ -46,6 +49,12 @@ import zope.testrunner.debug
 import zope.testrunner.tb_format
 import zope.testrunner.shuffle
 
+try:
+    import Queue # Python 2
+except ImportError:
+    # Python 3
+    import queue as Queue # Python 3
+
 try:
     from unittest import _UnexpectedSuccess # Python 3.1
 except ImportError:
@@ -87,7 +96,7 @@ class Runner(object):
     """
 
     def __init__(self, defaults=None, args=None, found_suites=None,
-                 options=None, script_parts=None):
+                 options=None, script_parts=None, cwd=None):
         if defaults is None:
             self.defaults = []
         else:
@@ -96,11 +105,14 @@ class Runner(object):
         self.found_suites = found_suites
         self.options = options
         self.script_parts = script_parts
+        self.cwd = cwd
         self.failed = True
 
         self.ran = 0
+        self.skipped = []
         self.failures = []
         self.errors = []
+        self.import_errors = []
 
         self.show_report = True
         self.do_run_tests = True
@@ -226,10 +238,11 @@ class Runner(object):
                 feature.layer_setup(layer)
             try:
                 self.ran += run_layer(self.options, layer_name, layer, tests,
-                                      setup_layers, self.failures, self.errors)
+                                      setup_layers, self.failures, self.errors,
+                                      self.skipped, self.import_errors)
             except zope.testrunner.interfaces.EndRun:
                 self.failed = True
-                return
+                break
             except CanNotTearDown:
                 if not self.options.resume_layer:
                     should_resume = True
@@ -244,7 +257,8 @@ class Runner(object):
             if layers_to_run:
                 self.ran += resume_tests(
                     self.script_parts, self.options, self.features,
-                    layers_to_run, self.failures, self.errors)
+                    layers_to_run, self.failures, self.errors,
+                    self.skipped, self.cwd)
 
         if setup_layers:
             if self.options.resume_layer is None:
@@ -254,7 +268,7 @@ class Runner(object):
         self.failed = bool(self.import_errors or self.failures or self.errors)
 
 
-def run_tests(options, tests, name, failures, errors):
+def run_tests(options, tests, name, failures, errors, skipped, import_errors):
     repeat = options.repeat or 1
     repeat_range = iter(range(repeat))
     ran = 0
@@ -268,7 +282,6 @@ def run_tests(options, tests, name, failures, errors):
         gc.collect()
         lgarbage = len(gc.garbage)
 
-    sumrc = 0
     if options.report_refcounts:
         if options.verbose:
             # XXX This code path is untested
@@ -326,9 +339,16 @@ def run_tests(options, tests, name, failures, errors):
             # Python versions prior to 2.7 do not have the concept of
             # unexpectedSuccesses.
             failures.extend(result.unexpectedSuccesses)
+        if not hasattr(result, 'skipped'):
+            # Only in Python >= 2.7, Python >= 3.1, and when using unittest2
+            result.skipped = []
+        skipped.extend(result.skipped)
         errors.extend(result.errors)
-        output.summary(result.testsRun, len(failures),
-            len(result.errors), t)
+        output.summary(n_tests=result.testsRun,
+                       n_failures=len(failures),
+                       n_errors=len(result.errors) + len(import_errors),
+                       n_seconds=t,
+                       n_skipped=len(result.skipped))
         ran = result.testsRun
 
         if is_jython:
@@ -363,7 +383,7 @@ def run_tests(options, tests, name, failures, errors):
 
 
 def run_layer(options, layer_name, layer, tests, setup_layers,
-              failures, errors):
+              failures, errors, skipped, import_errors):
 
     output = options.output
     gathered = []
@@ -381,13 +401,14 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
     except zope.testrunner.interfaces.EndRun:
         raise
     except Exception:
-        f = cStringIO.StringIO()
+        f = StringIO()
         traceback.print_exc(file=f)
         output.error(f.getvalue())
         errors.append((SetUpLayerFailure(layer), sys.exc_info()))
         return 0
     else:
-        return run_tests(options, tests, layer_name, failures, errors)
+        return run_tests(options, tests, layer_name, failures, errors, skipped,
+                         import_errors)
 
 
 class SetUpLayerFailure(unittest.TestCase):
@@ -405,14 +426,14 @@ class SetUpLayerFailure(unittest.TestCase):
 
 
 def spawn_layer_in_subprocess(result, script_parts, options, features,
-                              layer_name, layer, failures, errors,
-                              resume_number):
+                              layer_name, layer, failures, errors, skipped,
+                              resume_number, cwd=None):
     output = options.output
 
     try:
         # BBB
         if script_parts is None:
-            script_parts = sys.argv[0:1]
+            script_parts = zope.testrunner._script_parts()
         args = [sys.executable]
         args.extend(script_parts)
         args.extend(['--resume-layer', layer_name, str(resume_number)])
@@ -426,6 +447,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             and sys.version_info[:3] <= (2, 4, 1)):
             args.insert(1, '-O')
 
+        debugargs = args  # save them before messing up for windows
         if sys.platform.startswith('win'):
             args = args[0] + ' ' + ' '.join([
                 ('"' + a.replace('\\', '\\\\').replace('"', '\\"') + '"')
@@ -435,7 +457,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             feature.layer_setup(layer)
 
         child = subprocess.Popen(args, shell=False, stdin=subprocess.PIPE,
-            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
+            stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd,
             close_fds=not sys.platform.startswith('win'))
 
         while True:
@@ -451,7 +473,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                     if not l:
                         break
                     result.write(l)
-            except IOError, e:
+            except IOError as e:
                 if e.errno == errno.EINTR:
                     # If the subprocess dies before we finish reading its
                     # output, a SIGCHLD signal can interrupt the reading.
@@ -464,7 +486,8 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                 break
 
         # Now stderr should be ready to read the whole thing.
-        erriter = iter(child.stderr.read().splitlines())
+        errlines = child.stderr.read().splitlines()
+        erriter = iter(errlines)
         nfail = nerr = 0
         for line in erriter:
             try:
@@ -474,21 +497,38 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             else:
                 break
         else:
-            output.error_with_banner("Could not communicate with subprocess")
+            errmsg = "Could not communicate with subprocess!"
+            errors.append(("subprocess for %s" % layer_name, None))
+            if options.verbose >= 1:
+                errmsg += "\nChild command line: %s" % debugargs
+            if (options.verbose >= 2 or
+                (options.verbose == 1 and len(errlines) < 20)):
+                errmsg += ("\nChild stderr was:\n" +
+                           "\n".join("  " + line.decode('utf-8', 'replace')
+                                     for line in errlines))
+            elif options.verbose >= 1:
+                errmsg += ("\nChild stderr was:\n" +
+                           "\n".join("  " + line.decode('utf-8', 'replace')
+                                     for line in errlines[:10]) +
+                           "\n...\n" +
+                           "\n".join("  " + line.decode('utf-8', 'replace')
+                                     for line in errlines[-10:]))
+            output.error_with_banner(errmsg)
+
 
         while nfail > 0:
             nfail -= 1
             # Doing erriter.next().strip() confuses the 2to3 fixer, so
             # we need to do it on a separate line. Also, in python 3 this
             # returns bytes, so we decode it.
-            next_fail = erriter.next()
+            next_fail = next(erriter)
             failures.append((next_fail.strip().decode(), None))
         while nerr > 0:
             nerr -= 1
             # Doing erriter.next().strip() confuses the 2to3 fixer, so
             # we need to do it on a separate line. Also, in python 3 this
             # returns bytes, so we decode it.
-            next_err = erriter.next()
+            next_err = next(erriter)
             errors.append((next_err.strip().decode(), None))
 
     finally:
@@ -555,7 +595,8 @@ class KeepaliveSubprocessResult(AbstractSubprocessResult):
             self.stdout.append(out)
 
 
-def resume_tests(script_parts, options, features, layers, failures, errors):
+def resume_tests(script_parts, options, features, layers, failures, errors,
+                 skipped, cwd=None):
     results = []
     stdout_queue = None
     if options.processes == 1:
@@ -573,13 +614,13 @@ def resume_tests(script_parts, options, features, layers, failures, errors):
         ready_threads.append(threading.Thread(
             target=spawn_layer_in_subprocess,
             args=(result, script_parts, options, features, layer_name, layer,
-                  failures, errors, resume_number)))
+                  failures, errors, skipped, resume_number, cwd)))
         resume_number += 1
 
     # Now start a few threads at a time.
     running_threads = []
     results_iter = iter(results)
-    current_result = results_iter.next()
+    current_result = next(results_iter)
     last_layer_intermediate_output = None
     output = None
     while ready_threads or running_threads:
@@ -613,10 +654,10 @@ def resume_tests(script_parts, options, features, layers, failures, errors):
             if output is not None:
                 sys.stdout.write(']\n')
                 output = None
-            map(sys.stdout.write, current_result.stdout)
+            sys.stdout.writelines(current_result.stdout)
 
             try:
-                current_result = results_iter.next()
+                current_result = next(results_iter)
             except StopIteration:
                 current_result = None
 
@@ -738,6 +779,10 @@ class TestResult(unittest.TestResult):
         t = max(time.time() - self._start_time, 0.0)
         self.options.output.test_success(test, t)
 
+    def addSkip(self, test, reason):
+        unittest.TestResult.addSkip(self, test, reason)
+        self.options.output.test_skipped(test, reason)
+
     def addError(self, test, exc_info):
         self.options.output.test_error(test, time.time() - self._start_time,
                                        exc_info)
@@ -785,6 +830,7 @@ class TestResult(unittest.TestResult):
                                                       " when running a layer"
                                                       " as a subprocess!")
             else:
+                # XXX: what exc_info? there's no exc_info!
                 zope.testrunner.debug.post_mortem(exc_info)
         elif self.options.stop_on_error:
             self.stop()
@@ -826,7 +872,7 @@ def layer_from_name(layer_name):
     module = import_name(module_name)
     try:
         return getattr(module, module_layer_name)
-    except AttributeError, e:
+    except AttributeError:
         # the default error is very uninformative:
         #   AttributeError: 'module' object has no attribute 'DemoLayer'
         # it doesn't say *which* module
@@ -864,10 +910,10 @@ def gather_layers(layer, result):
 class FakeInputContinueGenerator:
 
     def readline(self):
-        print  'c\n'
-        print '*'*70
+        print('c\n')
+        print('*'*70)
         print ("Can't use pdb.set_trace when running a layer"
                " as a subprocess!")
-        print '*'*70
-        print
+        print('*'*70)
+        print()
         return 'c\n'
